@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtScxml module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "qscxmlcompiler_p.h"
 #include "qscxmlexecutablecontent_p.h"
@@ -43,7 +7,7 @@
 #include <qxmlstream.h>
 #include <qloggingcategory.h>
 #include <qfile.h>
-#include <qvector.h>
+#include <qlist.h>
 #include <qstring.h>
 
 #ifndef BUILD_QSCXMLC
@@ -56,6 +20,8 @@
 #include <private/qmetaobjectbuilder_p.h>
 #endif // BUILD_QSCXMLC
 
+#include <QtCore/qmap.h>
+
 #include <functional>
 
 namespace {
@@ -74,7 +40,7 @@ namespace {
 class ScxmlVerifier: public DocumentModel::NodeVisitor
 {
 public:
-    ScxmlVerifier(std::function<void (const DocumentModel::XmlLocation &, const QString &)> errorHandler)
+    ScxmlVerifier(std::function<void (const QString &, const DocumentModel::XmlLocation &, const QString &)> errorHandler)
         : m_errorHandler(errorHandler)
         , m_doc(nullptr)
         , m_hasErrors(false)
@@ -87,7 +53,7 @@ public:
 
         doc->isVerified = true;
         m_doc = doc;
-        for (DocumentModel::AbstractState *state : qAsConst(doc->allStates)) {
+        for (DocumentModel::AbstractState *state : std::as_const(doc->allStates)) {
             if (state->id.isEmpty()) {
                 continue;
 #ifndef QT_NO_DEBUG
@@ -117,8 +83,8 @@ private:
                 scxml->initialTransition = createInitialTransition({firstChild});
             }
         } else {
-            QVector<DocumentModel::AbstractState *> initialStates;
-            for (const QString &initial : qAsConst(scxml->initial)) {
+            QList<DocumentModel::AbstractState *> initialStates;
+            for (const QString &initial : std::as_const(scxml->initial)) {
                 if (DocumentModel::AbstractState *s = m_stateById.value(initial))
                     initialStates.append(s);
                 else
@@ -155,8 +121,8 @@ private:
                 }
             } else {
                 Q_ASSERT(state->type == DocumentModel::State::Normal);
-                QVector<DocumentModel::AbstractState *> initialStates;
-                for (const QString &initialState : qAsConst(state->initial)) {
+                QList<DocumentModel::AbstractState *> initialStates;
+                for (const QString &initialState : std::as_const(state->initial)) {
                     if (DocumentModel::AbstractState *s = m_stateById.value(initialState)) {
                         initialStates.append(s);
                     } else {
@@ -207,7 +173,7 @@ private:
 
         if (int size = transition->targets.size())
             transition->targetStates.reserve(size);
-        for (const QString &target : qAsConst(transition->targets)) {
+        for (const QString &target : std::as_const(transition->targets)) {
             if (DocumentModel::AbstractState *s = m_stateById.value(target)) {
                 if (transition->targetStates.contains(s)) {
                     error(transition->xmlLocation, QStringLiteral("duplicate target '%1'").arg(target));
@@ -218,7 +184,7 @@ private:
                 error(transition->xmlLocation, QStringLiteral("unknown state '%1' in target").arg(target));
             }
         }
-        for (const QString &event : qAsConst(transition->events))
+        for (const QString &event : std::as_const(transition->events))
             checkEvent(event, transition->xmlLocation, AllowWildCards);
 
         m_parentNodes.append(transition);
@@ -233,7 +199,7 @@ private:
     bool visit(DocumentModel::HistoryState *state) override
     {
         bool seenTransition = false;
-        for (DocumentModel::StateOrTransition *sot : qAsConst(state->children)) {
+        for (DocumentModel::StateOrTransition *sot : std::as_const(state->children)) {
             if (DocumentModel::State *s = sot->asState()) {
                 error(s->xmlLocation, QStringLiteral("history state cannot have substates"));
             } else if (DocumentModel::Transition *t = sot->asTransition()) {
@@ -298,7 +264,7 @@ private:
             if (!isLetter(c) && c != QLatin1Char('_'))
                 return false;
         }
-        for (int ei = id.length(); i != ei; ++i) {
+        for (int ei = id.size(); i != ei; ++i) {
             const QChar c = id.at(i);
             if (isLetter(c) || c.isDigit() || c == QLatin1Char('.') || c == QLatin1Char('-')
                     || c == QLatin1Char('_') || isNameTail(c))
@@ -370,12 +336,12 @@ private:
             if (part.isEmpty())
                 return false;
 
-            if (wildCardMode == AllowWildCards && part.length() == 1
+            if (wildCardMode == AllowWildCards && part.size() == 1
                     && part.at(0) == QLatin1Char('*')) {
                 continue;
             }
 
-            for (int i = 0, ei = part.length(); i != ei; ++i) {
+            for (int i = 0, ei = part.size(); i != ei; ++i) {
                 const QChar c = part.at(i);
                 if (!isLetter(c) && !c.isDigit() && c != QLatin1Char('-') && c != QLatin1Char('_')
                         && c != QLatin1Char(':')) {
@@ -387,7 +353,7 @@ private:
         return true;
     }
 
-    static const QVector<DocumentModel::StateOrTransition *> &allChildrenOfContainer(
+    static const QList<DocumentModel::StateOrTransition *> &allChildrenOfContainer(
             DocumentModel::StateContainer *container)
     {
         if (auto state = container->asState())
@@ -402,8 +368,8 @@ private:
     {
         const auto &allChildren = allChildrenOfContainer(container);
 
-        QVector<DocumentModel::AbstractState *> childStates;
-        for (DocumentModel::StateOrTransition *child : qAsConst(allChildren)) {
+        QList<DocumentModel::AbstractState *> childStates;
+        for (DocumentModel::StateOrTransition *child : std::as_const(allChildren)) {
             if (DocumentModel::State *s = child->asState())
                 return s;
             else if (DocumentModel::HistoryState *h = child->asHistoryState())
@@ -412,13 +378,13 @@ private:
         return nullptr;
     }
 
-    static QVector<DocumentModel::AbstractState *> allAbstractStates(
+    static QList<DocumentModel::AbstractState *> allAbstractStates(
             DocumentModel::StateContainer *container)
     {
         const auto &allChildren = allChildrenOfContainer(container);
 
-        QVector<DocumentModel::AbstractState *> childStates;
-        for (DocumentModel::StateOrTransition *child : qAsConst(allChildren)) {
+        QList<DocumentModel::AbstractState *> childStates;
+        for (DocumentModel::StateOrTransition *child : std::as_const(allChildren)) {
             if (DocumentModel::State *s = child->asState())
                 childStates.append(s);
             else if (DocumentModel::HistoryState *h = child->asHistoryState())
@@ -428,7 +394,7 @@ private:
     }
 
     DocumentModel::Transition *createInitialTransition(
-            const QVector<DocumentModel::AbstractState *> &states)
+            const QList<DocumentModel::AbstractState *> &states)
     {
         auto *newTransition = m_doc->newTransition(nullptr, DocumentModel::XmlLocation(-1, -1));
         newTransition->type = DocumentModel::Transition::Synthetic;
@@ -452,15 +418,15 @@ private:
     {
         m_hasErrors = true;
         if (m_errorHandler)
-            m_errorHandler(location, message);
+            m_errorHandler(m_doc->fileName, location, message);
     }
 
 private:
-    std::function<void (const DocumentModel::XmlLocation &, const QString &)> m_errorHandler;
+    std::function<void (const QString &, const DocumentModel::XmlLocation &, const QString &)> m_errorHandler;
     DocumentModel::ScxmlDocument *m_doc;
     bool m_hasErrors;
     QHash<QString, DocumentModel::AbstractState *> m_stateById;
-    QVector<DocumentModel::Node *> m_parentNodes;
+    QList<DocumentModel::Node *> m_parentNodes;
 };
 
 #ifndef BUILD_QSCXMLC
@@ -469,8 +435,8 @@ class InvokeDynamicScxmlFactory: public QScxmlInvokableServiceFactory
     Q_OBJECT
 public:
     InvokeDynamicScxmlFactory(const QScxmlExecutableContent::InvokeInfo &invokeInfo,
-                              const QVector<QScxmlExecutableContent::StringId> &namelist,
-                              const QVector<QScxmlExecutableContent::ParameterInfo> &params)
+                              const QList<QScxmlExecutableContent::StringId> &namelist,
+                              const QList<QScxmlExecutableContent::ParameterInfo> &params)
         : QScxmlInvokableServiceFactory(invokeInfo, namelist, params)
     {}
 
@@ -487,7 +453,11 @@ class DynamicStateMachinePrivate : public QScxmlStateMachinePrivate
 {
     struct DynamicMetaObject : public QAbstractDynamicMetaObject
     {
-        QAbstractDynamicMetaObject *toDynamicMetaObject(QObject *) override
+#if QT_VERSION >= QT_VERSION_CHECK(7, 0, 0)
+        const QMetaObject *toDynamicMetaObject(QObject *) const override
+#else
+        QMetaObject *toDynamicMetaObject(QObject *) override
+#endif
         {
             return this;
         }
@@ -624,8 +594,8 @@ public:
         DataModelInfo dm;
         auto factoryIdCreator = [stateMachine](
                 const QScxmlExecutableContent::InvokeInfo &invokeInfo,
-                const QVector<QScxmlExecutableContent::StringId> &namelist,
-                const QVector<QScxmlExecutableContent::ParameterInfo> &params,
+                const QList<QScxmlExecutableContent::StringId> &namelist,
+                const QList<QScxmlExecutableContent::ParameterInfo> &params,
                 const QSharedPointer<DocumentModel::ScxmlDocument> &content) -> int {
             auto factory = new InvokeDynamicScxmlFactory(invokeInfo, namelist, params);
             factory->setContent(content);
@@ -651,7 +621,7 @@ private:
     }
 
 private:
-    QVector<QScxmlInvokableServiceFactory *> m_allFactoriesById;
+    QList<QScxmlInvokableServiceFactory *> m_allFactoriesById;
     int m_propertyCount;
 };
 
@@ -862,7 +832,7 @@ QScxmlStateMachine *QScxmlCompilerPrivate::instantiateStateMachine() const
 void QScxmlCompilerPrivate::instantiateDataModel(QScxmlStateMachine *stateMachine) const
 {
 #ifdef BUILD_QSCXMLC
-    Q_UNUSED(stateMachine)
+    Q_UNUSED(stateMachine);
 #else
     if (!m_errors.isEmpty()) {
         qWarning() << "SCXML document has errors";
@@ -886,7 +856,7 @@ void QScxmlCompilerPrivate::instantiateDataModel(QScxmlStateMachine *stateMachin
 /*!
  * Returns the list of parse errors.
  */
-QVector<QScxmlError> QScxmlCompiler::errors() const
+QList<QScxmlError> QScxmlCompiler::errors() const
 {
     return d->errors();
 }
@@ -1030,7 +1000,7 @@ bool QScxmlCompilerPrivate::ParserState::isExecutableContent(ParserState::Kind k
     return false;
 }
 
-QScxmlCompilerPrivate::ParserState::Kind QScxmlCompilerPrivate::ParserState::nameToParserStateKind(const QStringRef &name)
+QScxmlCompilerPrivate::ParserState::Kind QScxmlCompilerPrivate::ParserState::nameToParserStateKind(QStringView name)
 {
     static QMap<QString, ParserState::Kind> nameToKind;
     if (nameToKind.isEmpty()) {
@@ -1195,7 +1165,7 @@ void DocumentModel::Param::accept(DocumentModel::NodeVisitor *visitor)
 void DocumentModel::DoneData::accept(DocumentModel::NodeVisitor *visitor)
 {
     if (visitor->visit(this)) {
-        for (Param *param : qAsConst(params))
+        for (Param *param : std::as_const(params))
             param->accept(visitor);
     }
     visitor->endVisit(this);
@@ -1268,7 +1238,7 @@ void DocumentModel::State::accept(DocumentModel::NodeVisitor *visitor)
         visitor->visit(onExit);
         if (doneData)
             doneData->accept(visitor);
-        for (Invoke *invoke : qAsConst(invokes))
+        for (Invoke *invoke : std::as_const(invokes))
             invoke->accept(visitor);
     }
     visitor->endVisit(this);
@@ -1351,11 +1321,11 @@ bool QScxmlCompilerPrivate::verifyDocument()
     if (!m_doc)
         return false;
 
-    auto handler = [this](const DocumentModel::XmlLocation &location, const QString &msg) {
-        this->addError(location, msg);
+    auto handler = [this](const QString &fileName, const DocumentModel::XmlLocation &location, const QString &msg) {
+        this->addError(fileName, location, msg);
     };
 
-    if (ScxmlVerifier(handler).verify(m_doc.data()))
+    if (ScxmlVerifier(handler).verify(m_doc.get()))
         return true;
     else
         return false;
@@ -1363,7 +1333,7 @@ bool QScxmlCompilerPrivate::verifyDocument()
 
 DocumentModel::ScxmlDocument *QScxmlCompilerPrivate::scxmlDocument() const
 {
-    return m_doc && m_errors.isEmpty() ? m_doc.data() : nullptr;
+    return m_doc && m_errors.isEmpty() ? m_doc.get() : nullptr;
 }
 
 QString QScxmlCompilerPrivate::fileName() const
@@ -1394,7 +1364,7 @@ void QScxmlCompilerPrivate::parseSubDocument(DocumentModel::Invoke *parentInvoke
     p.setFileName(fileName);
     p.setLoader(loader());
     p.d->readDocument();
-    parentInvoke->content.reset(p.d->m_doc.take());
+    parentInvoke->content.reset(p.d->m_doc.release());
     m_doc->allSubDocuments.append(parentInvoke->content.data());
     m_errors.append(p.errors());
 }
@@ -1408,7 +1378,7 @@ bool QScxmlCompilerPrivate::parseSubElement(DocumentModel::Invoke *parentInvoke,
     p.setLoader(loader());
     p.d->resetDocument();
     bool ok = p.d->readElement();
-    parentInvoke->content.reset(p.d->m_doc.take());
+    parentInvoke->content.reset(p.d->m_doc.release());
     m_doc->allSubDocuments.append(parentInvoke->content.data());
     m_errors.append(p.errors());
     return ok;
@@ -1429,7 +1399,7 @@ bool QScxmlCompilerPrivate::preReadElementScxml()
         scxml->initial += initial.split(QChar::Space, Qt::SkipEmptyParts);
     }
 
-    const QStringRef datamodel = attributes.value(QLatin1String("datamodel"));
+    const QStringView datamodel = attributes.value(QLatin1String("datamodel"));
     if (datamodel.isEmpty() || datamodel == QLatin1String("null")) {
         scxml->dataModel = DocumentModel::Scxml::NullDataModel;
     } else if (datamodel == QLatin1String("ecmascript")) {
@@ -1443,7 +1413,7 @@ bool QScxmlCompilerPrivate::preReadElementScxml()
         } else {
             int lastColon = datamodel.lastIndexOf(QLatin1Char(':'));
             if (lastColon == -1) {
-                lastColon = datamodel.length();
+                lastColon = datamodel.size();
             } else {
                 scxml->cppDataModelHeaderName = datamodel.mid(lastColon + 1).toString();
             }
@@ -1453,7 +1423,7 @@ bool QScxmlCompilerPrivate::preReadElementScxml()
         addError(QStringLiteral("Unsupported data model '%1' in scxml")
                  .arg(datamodel.toString()));
     }
-    const QStringRef binding = attributes.value(QLatin1String("binding"));
+    const QStringView binding = attributes.value(QLatin1String("binding"));
     if (binding.isEmpty() || binding == QLatin1String("early")) {
         scxml->binding = DocumentModel::Scxml::EarlyBinding;
     } else if (binding == QLatin1String("late")) {
@@ -1463,7 +1433,7 @@ bool QScxmlCompilerPrivate::preReadElementScxml()
                  .arg(binding.toString()));
         return false;
     }
-    const QStringRef name = attributes.value(QLatin1String("name"));
+    const QStringView name = attributes.value(QLatin1String("name"));
     if (!name.isEmpty()) {
         scxml->name = name.toString();
     }
@@ -1551,7 +1521,7 @@ bool QScxmlCompilerPrivate::preReadElementTransition()
     transition->targets = attributes.value(QLatin1String("target")).toString().split(QLatin1Char(' '), Qt::SkipEmptyParts);
     if (attributes.hasAttribute(QStringLiteral("cond")))
         transition->condition.reset(new QString(attributes.value(QLatin1String("cond")).toString()));
-    QStringRef type = attributes.value(QLatin1String("type"));
+    QStringView type = attributes.value(QLatin1String("type"));
     if (type.isEmpty() || type == QLatin1String("external")) {
         transition->type = DocumentModel::Transition::External;
     } else if (type == QLatin1String("internal")) {
@@ -1587,7 +1557,7 @@ bool QScxmlCompilerPrivate::preReadElementHistory()
     if (!maybeId(attributes, &newState->id))
         return false;
 
-    const QStringRef type = attributes.value(QLatin1String("type"));
+    const QStringView type = attributes.value(QLatin1String("type"));
     if (type.isEmpty() || type == QLatin1String("shallow")) {
         newState->type = DocumentModel::HistoryState::Shallow;
     } else if (type == QLatin1String("deep")) {
@@ -1867,11 +1837,11 @@ bool QScxmlCompilerPrivate::preReadElementInvoke()
     invoke->idLocation = attributes.value(QLatin1String("idlocation")).toString();
     invoke->type = attributes.value(QLatin1String("type")).toString();
     invoke->typeexpr = attributes.value(QLatin1String("typeexpr")).toString();
-    QStringRef autoforwardS = attributes.value(QLatin1String("autoforward"));
-    if (QStringRef::compare(autoforwardS, QLatin1String("true"), Qt::CaseInsensitive) == 0
-            || QStringRef::compare(autoforwardS, QLatin1String("yes"), Qt::CaseInsensitive) == 0
-            || QStringRef::compare(autoforwardS, QLatin1String("t"), Qt::CaseInsensitive) == 0
-            || QStringRef::compare(autoforwardS, QLatin1String("y"), Qt::CaseInsensitive) == 0
+    QStringView autoforwardS = attributes.value(QLatin1String("autoforward"));
+    if (autoforwardS.compare(QLatin1String("true"), Qt::CaseInsensitive) == 0
+            || autoforwardS.compare(QLatin1String("yes"), Qt::CaseInsensitive) == 0
+            || autoforwardS.compare(QLatin1String("t"), Qt::CaseInsensitive) == 0
+            || autoforwardS.compare(QLatin1String("y"), Qt::CaseInsensitive) == 0
             || autoforwardS == QLatin1String("1"))
         invoke->autoforward = true;
     else
@@ -2082,6 +2052,9 @@ bool QScxmlCompilerPrivate::postReadElementScript()
                 scriptI->content = QString::fromUtf8(data);
             }
         }
+    } else {
+        addError(m_doc->fileName, scriptI->xmlLocation,
+                QStringLiteral("neither src nor any content has been given in the script tag"));
     }
     return flushInstruction();
 }
@@ -2135,7 +2108,7 @@ bool QScxmlCompilerPrivate::readDocument()
     for (bool finished = false; !finished && !m_reader->hasError();) {
         switch (m_reader->readNext()) {
         case QXmlStreamReader::StartElement : {
-            const QStringRef newTag = m_reader->name();
+            const QStringView newTag = m_reader->name();
             const ParserState::Kind newElementKind = ParserState::nameToParserStateKind(newTag);
 
             auto ns = m_reader->namespaceUri();
@@ -2176,7 +2149,7 @@ bool QScxmlCompilerPrivate::readDocument()
 
 bool QScxmlCompilerPrivate::readElement()
 {
-    const QStringRef currentTag = m_reader->name();
+    const QStringView currentTag = m_reader->name();
     const QXmlStreamAttributes attributes = m_reader->attributes();
 
     const ParserState::Kind elementKind = ParserState::nameToParserStateKind(currentTag);
@@ -2199,7 +2172,7 @@ bool QScxmlCompilerPrivate::readElement()
         return parseSubElement(i, m_reader, m_fileName);
     }
 
-    if (elementKind != ParserState::Scxml && !m_stack.count()) {
+    if (elementKind != ParserState::Scxml && !m_stack.size()) {
         addError(QStringLiteral("misplaced %1").arg(currentTag.toString()));
         return false;
     }
@@ -2241,7 +2214,7 @@ bool QScxmlCompilerPrivate::readElement()
     for (bool finished = false; !finished && !m_reader->hasError();) {
         switch (m_reader->readNext()) {
         case QXmlStreamReader::StartElement : {
-            const QStringRef newTag = m_reader->name();
+            const QStringView newTag = m_reader->name();
             const ParserState::Kind newElementKind = ParserState::nameToParserStateKind(newTag);
 
             auto ns = m_reader->namespaceUri();
@@ -2349,7 +2322,7 @@ QByteArray QScxmlCompilerPrivate::load(const QString &name, bool *ok)
     return result;
 }
 
-QVector<QScxmlError> QScxmlCompilerPrivate::errors() const
+QList<QScxmlError> QScxmlCompilerPrivate::errors() const
 {
     return m_errors;
 }
@@ -2359,9 +2332,9 @@ void QScxmlCompilerPrivate::addError(const QString &msg)
     m_errors.append(QScxmlError(m_fileName, m_reader->lineNumber(), m_reader->columnNumber(), msg));
 }
 
-void QScxmlCompilerPrivate::addError(const DocumentModel::XmlLocation &location, const QString &msg)
+void QScxmlCompilerPrivate::addError(const QString &fileName, const DocumentModel::XmlLocation &location, const QString &msg)
 {
-    m_errors.append(QScxmlError(m_fileName, location.line, location.column, msg));
+    m_errors.append(QScxmlError(fileName, location.line, location.column, msg));
 }
 
 DocumentModel::AbstractState *QScxmlCompilerPrivate::currentParent() const
@@ -2380,7 +2353,7 @@ bool QScxmlCompilerPrivate::maybeId(const QXmlStreamAttributes &attributes, QStr
     QString idStr = attributes.value(QLatin1String("id")).toString();
     if (!idStr.isEmpty()) {
         if (m_allIds.contains(idStr)) {
-            addError(xmlLocation(), QStringLiteral("duplicate id '%1'").arg(idStr));
+            addError(m_doc->fileName, xmlLocation(), QStringLiteral("duplicate id '%1'").arg(idStr));
         } else {
             m_allIds.insert(idStr);
             *id = idStr;
@@ -2416,12 +2389,12 @@ QScxmlCompilerPrivate::ParserState &QScxmlCompilerPrivate::current()
 
 QScxmlCompilerPrivate::ParserState &QScxmlCompilerPrivate::previous()
 {
-    return m_stack[m_stack.count() - 2];
+    return m_stack[m_stack.size() - 2];
 }
 
 bool QScxmlCompilerPrivate::hasPrevious() const
 {
-    return m_stack.count() > 1;
+    return m_stack.size() > 1;
 }
 
 bool QScxmlCompilerPrivate::checkAttributes(const QXmlStreamAttributes &attributes,
@@ -2438,7 +2411,7 @@ bool QScxmlCompilerPrivate::checkAttributes(const QXmlStreamAttributes &attribut
 {
     QStringList required = requiredNames;
     for (const QXmlStreamAttribute &attribute : attributes) {
-        const QStringRef ns = attribute.namespaceUri();
+        const QStringView ns = attribute.namespaceUri();
         if (!ns.isEmpty() && ns != scxmlNamespace && ns != qtScxmlNamespace)
             continue;
 
@@ -2473,7 +2446,7 @@ QByteArray QScxmlCompilerPrivate::DefaultLoader::load(const QString &name, const
     const QUrl url(name);
     if (!url.isLocalFile() && !url.isRelative())
         errs << QStringLiteral("src attribute is not a local file (%1)").arg(name);
-    QFileInfo fInfo = url.isLocalFile() ? url.toLocalFile() : name;
+    QFileInfo fInfo(url.isLocalFile() ? url.toLocalFile() : name);
 #endif // BUILD_QSCXMLC
     if (fInfo.isRelative())
         fInfo = QFileInfo(QDir(baseDir).filePath(fInfo.filePath()));
